from ..entities.grammar_tree_nodes import RootNode, InterpolatedNode, FixedNode
from ..entities.wrappers import KCScalar

def parse_grammar(value):
    root_node = RootNode()
    _parse(value, root_node)
    return root_node


def _parse(value, parent_node):
    if isinstance(value, KCScalar):
        value = value.value
    if not isinstance(value, str):
        raise TypeError

    while True:
        if len(value) == 0:
            return
        if "${" not in value:
            parent_node.children.append(FixedNode(value))
            return

        brace_start = value.index("${")
        if brace_start != 0:
            parent_node.children.append(FixedNode(value[:brace_start]))
        value = value[brace_start + 2:]
        brace_end = _find_brace_end(value)

        value_in_braces = value[:brace_end]
        if ":" in value_in_braces:
            colon_idx = value_in_braces.index(":")
            if colon_idx == 0:
                raise ValueError
            else:
                node = InterpolatedNode(value_in_braces[:colon_idx])
        else:
            node = InterpolatedNode(None)
            colon_idx = -1
        parent_node.children.append(node)

        # parse recursively
        _parse(value[colon_idx + 1:brace_end], node)
        value = value[brace_end + 1:]


def _find_brace_end(value):
    nesting_level = 1
    for i in range(len(value)):
        if value[i] == "}":
            nesting_level -= 1
            if nesting_level == 0:
                return i
        if value[i] == "$" and i != len(value) - 2 and value[i + 1] == "{":
            nesting_level += 1
    raise ValueError


from .resolver import Resolver
from .scalar_resolvers.interpolation_resolver import InterpolationResolver
from .scalar_resolvers.eval_resolver import EvalResolver
from .collection_resolvers.template_resolver import TemplateResolver
from .scalar_resolvers.nested_yaml_resolver import NestedYamlResolver

class DefaultResolver(Resolver):
    def __init__(self, template_path=None, **templates):
        super().__init__(
            TemplateResolver(template_path=template_path, **(templates or {})),
            default_scalar_resolver=InterpolationResolver(),
            eval=EvalResolver(),
            yaml=NestedYamlResolver(template_path=template_path, **(templates or {})),
        )


class KCObject:
    def __repr__(self):
        raise NotImplementedError

    def __str__(self):
        return repr(self)

class KCScalar(KCObject):
    def __init__(self, value):
        super().__init__()
        self.value = value

    def __repr__(self):
        return repr(self.value)

class KCDict(KCObject):
    def __init__(self, **kwargs):
        super().__init__()
        self.dict = kwargs

    def __repr__(self):
        return repr(self.dict)

    def __setitem__(self, key, value):
        self.dict[key] = value

    def __getitem__(self, item):
        return self.dict[item]

    def __contains__(self, item):
        return item in self.dict

    def keys(self):
        return self.dict.keys()

    def items(self):
        return self.dict.items()

class KCList(KCObject):
    def __init__(self, *args):
        super().__init__()
        self.list = list(args)

    def __repr__(self):
        return repr(self.list)

    def _check_accessor(self, accessor):
        if not isinstance(accessor, int):
            raise ValueError(f"expected int argument but got '{accessor}'")
        if not 0 <= accessor < len(self.list):
            raise IndexError(f"index {accessor} is out of range (0<={accessor}<{len(self.list)} evaluates to false)")

    def __setitem__(self, key, value):
        self._check_accessor(key)
        self.list[key] = value

    def __getitem__(self, item):
        return self.list[item]



from .scalar_resolver import ScalarResolver
from ...functional.util import string_to_accessors, select

class InterpolationResolver(ScalarResolver):
    def resolve(self, value, root_node):
        accessors = string_to_accessors(value)
        node = select(root_node=root_node, accessors=accessors)
        return node.value


from ..entities.wrappers import KCDict, KCList, KCScalar, KCObject
from ..functional.parse_grammar import parse_grammar
from ..entities.grammar_tree_nodes import RootNode, FixedNode, InterpolatedNode
from ..functional.util import string_to_accessors, select

class Resolver:
    def __init__(self, *collection_resolvers, default_scalar_resolver=None, **scalar_resolvers):
        self.collection_resolvers = list(collection_resolvers)
        self.scalar_resolvers = scalar_resolvers
        if default_scalar_resolver is not None:
            self.scalar_resolvers[None] = default_scalar_resolver


    def resolve(self, node, root_node=None):
        result = {}
        if root_node is None:
            root_node = node
        wrapped_node = KCDict(root=node)
        self._resolve_collection(node, root_node=root_node, result=result, trace=[(wrapped_node, "root")])
        return result["root"]


    def _resolve_collection(self, node, root_node, result, trace):
        if isinstance(node, KCDict):
            parent, parent_accessor = trace[-1]
            result[parent_accessor] = {}
            # preorder
            for resolver in self.collection_resolvers:
                resolver.preorder_resolve(node, root_node=root_node, result=result, trace=trace, root_resolver=self)
                # preorder is allowed to change the current node
                node = parent[parent_accessor]

            # traverse
            for accessor, subnode in node.dict.items():
                trace.append((node, accessor))
                self._resolve_collection(subnode, root_node=root_node, result=result[parent_accessor], trace=trace)
                trace.pop()

            # postorder
            for resolver in self.collection_resolvers:
                resolver.postorder_resolve(node, root_node=root_node, result=result, trace=trace, root_resolver=self)
        elif isinstance(node, KCList):
            parent, parent_accessor = trace[-1]
            result[parent_accessor] = []
            # preorder
            for resolver in self.collection_resolvers:
                resolver.preorder_resolve(node, root_node=root_node, result=result, trace=trace, root_resolver=self)
                # preorder is allowed to change the current node
                node = parent[parent_accessor]

            # traverse
            for i, subnode in enumerate(node.list):
                trace.append((subnode, i))
                self._resolve_collection(subnode, root_node=root_node, result=result[parent_accessor], trace=trace)
                trace.pop()

            # postorder
            for resolver in self.collection_resolvers:
                resolver.postorder_resolve(node, root_node=root_node, result=result, trace=trace, root_resolver=self)
        elif isinstance(node, KCScalar):
            if not (isinstance(result, list) or isinstance(result, dict)):
                raise TypeError
            if not isinstance(node.value, str):
                resolve_result = node.value
            else:
                resolve_result = self.resolve_scalar(node.value, root_node=root_node)

            # resolved value might be a KCObject (e.g. when loading a nested yaml)
            if isinstance(resolve_result, KCObject):
                parent, accessor = trace.pop()
                trace.append((resolve_result, accessor))
                self._resolve_collection(resolve_result, root_node=root_node, result=result, trace=trace)
            else:
                # set value
                parent_accessor = trace[-1][1]
                if isinstance(parent_accessor, int):
                    if len(result) != parent_accessor:
                        raise IndexError
                    result.append(resolve_result)
                else:
                    result[parent_accessor] = resolve_result
        else:
            raise TypeError

    def resolve_scalar(self, value, root_node):
        grammar_tree = parse_grammar(value)
        return self._resolve_scalar(grammar_tree, root_node=root_node)

    def _resolve_scalar(self, grammar_node, root_node):
        if isinstance(grammar_node, RootNode):
            resolve_results = [self._resolve_scalar(child, root_node=root_node) for child in grammar_node.children]
            return self._merge_scalar_resolve_results(resolve_results)
        elif isinstance(grammar_node, FixedNode):
            return grammar_node.value
        elif isinstance(grammar_node, InterpolatedNode):
            # resolve children
            resolve_results = [self._resolve_scalar(child, root_node=root_node) for child in grammar_node.children]
            resolve_result = self._merge_scalar_resolve_results(resolve_results)
            # resolve cur node
            scalar_resolver = self.scalar_resolvers[grammar_node.resolver_key]
            return scalar_resolver.resolve(resolve_result, root_node=root_node)
        else:
            raise TypeError

    @staticmethod
    def _merge_scalar_resolve_results(resolve_results):
        if len(resolve_results) == 1:
            # keep datatype of single result
            return resolve_results[0]
        else:
            # concat as string
            return "".join(map(str, resolve_results))


from .util import apply
from copy import deepcopy
from ..entities.wrappers import KCDict, KCList, KCScalar

def from_primitive(root_node):
    root_node = deepcopy(root_node)
    wrapped_node = dict(root=root_node)
    apply(root_node, parent_node=wrapped_node, parent_accessor="root", pre_fn=_from_primitive_fn)
    return wrapped_node["root"]

def _from_primitive_fn(node, parent_node, parent_accessor, **_):
    if isinstance(node, dict):
        parent_node[parent_accessor] = KCDict(**node)
    elif isinstance(node, list):
        parent_node[parent_accessor] = KCList(*node)
    elif not isinstance(node, (KCDict, KCList, KCScalar)):
        parent_node[parent_accessor] = KCScalar(node)



from .scalar_resolver import ScalarResolver
from ...entities.wrappers import KCObject, KCScalar
from ...functional.load import from_string, from_file_uri

class NestedYamlResolver(ScalarResolver):
    def __init__(self, template_path=None, **templates):
        super().__init__()
        self.template_path = template_path
        self.templates = templates
        if any(map(lambda template_value: not isinstance(template_value, str), templates.values())):
            raise TypeError

    def resolve(self, value, *_, **__):
        if isinstance(value, KCScalar):
            value = value.value
        if not isinstance(value, str) or not value.endswith(".yaml"):
            raise ValueError

        if value in self.templates:
            # load from templates dict (pretty much only used for testing)
            template = from_string(self.templates[value])
        else:
            # load from template_path
            template = from_file_uri(self.template_path / value)

        return template